@@ -1,59 +1,25 @@
 # -*- coding: utf-8 -*-
 """Different techniques for fitting baselines to experimental data.
 
- Morphological
-    1) mpls (Morphological Penalized Least Squares)
-    2) mor (Morphological)
-    3) imor (Improved Morphological)
-    4) iamor (Iterative averaging morphological)
-
 Created on March 5, 2021
 @author: Donald Erb
 
 """
 
 import numpy as np
-from scipy.ndimage import grey_closing, grey_dilation, grey_erosion, grey_opening
-from scipy.sparse.linalg import spsolve
+from scipy.linalg import solveh_banded
+from scipy.ndimage import grey_closing, grey_dilation, grey_erosion, grey_opening, uniform_filter1d
 
-from .utils import _setup_pls, mollify, relative_difference
+from . import utils
+from ._algorithm_setup import _optimize_window, _setup_morphology, _setup_whittaker
+from .utils import _HAS_PENTAPY, _pentapy_solve, pad_edges, padded_convolve, relative_difference
 
 
-def _smooth(data, x_data=None, use_whittaker=True, lam=1e6,
-            order=2, weights=None):
+# make _optimize_window available to users and document it so that
+# it is included in the api documentation
+def optimize_window(*args, **kwargs):
     """
-    Smooth a baseline using either Whittaker smoothing.
-
-    Parameters
-    ----------
-    data : [type]
-        [description]
-    x_data : [type], optional
-        [description], by default None
-    use_whittaker : bool, optional
-        [description], by default True
-    lam : [type], optional
-        [description], by default 1e6
-    order : int, optional
-        [description], by default 2
-
-    Returns
-    -------
-    [type]
-        [description]
-    """
-    y = np.asarray(data)
-    if use_whittaker:
-        _, D, W, w = _setup_pls(y, lam, order, weights)
-        z = spsolve(W + D, w * y)
-    else:
-        raise ValueError  #TODO need any other smoothers? maybe add the mollify/convolution here
-    return z
-
-
-def optimize_window(data, increment=1, max_hits=1, window_tol=1e-6, max_half_window=None):
-    """
-    Optimizes the morphological window size.
+    Optimizes the morphological half-window size.
 
     Parameters
     ----------
@@ -64,112 +30,61 @@ def optimize_window(data, increment=1, max_hits=1, window_tol=1e-6, max_half_win
     max_hits : int, optional
         The number of consecutive half windows that must produce the same
         morphological opening before accepting the half window as the optimum
-        value. Default is 1.
+        value. Default is 3.
     window_tol : float, optional
         The tolerance value for considering two morphological openings as
         equivalent. Default is 1e-6.
     max_half_window : int, optional
-        The maximum allowable window size. If None (default), will be set
+        The maximum allowable half-window size. If None (default), will be set
         to (len(data) - 1) / 2.
+    min_half_window : int, optional
+        The minimum half-window size. If None (default), will be set to 1.
 
     Returns
     -------
     half_window : int
         The optimized half window size.
 
+    Notes
+    -----
+    May only provide good results for some morphological algorithms, so use with
+    caution.
+
     References
     ----------
-    Dai, L., et al.. An Automated Baseline Correction Method Based on Iterative
-    Morphological Operations. Applied Spectroscopy, 2018, 72(5), 731-739.
+    Perez-Pueyo, R., et al. Morphology-Based Automated Baseline Removal for
+    Raman Spectra of Artistic Pigments. Applied Spectroscopy, 2010, 64, 595-600.
 
+    """
+    return _optimize_window(*args, **kwargs)
+
+
+def _mollifier_kernel(window_size):
+    """
+    A kernel for smoothing/mollification.
+
+    Parameters
+    ----------
+    window_size : int
+        The number of points for the entire kernel.
+
+    Returns
+    -------
+    numpy.ndarray, shape (2 * window_size + 1,)
+        The area normalized kernel.
+
+    References
+    ----------
     Chen, H., et al. An Adaptive and Fully Automated Baseline Correction
     Method for Raman Spectroscopy Based on Morphological Operations and
     Mollifications. Applied Spectroscopy, 2019, 73(3), 284-293.
 
     """
-    y = np.asarray(data)
-    if max_half_window is None:
-        max_half_window = (y.shape[0] - 1) / 2
-
-    opening = grey_opening(y, [3])  # half window = 1 to start
-    hits = 0
-    best_half_window = 1
-    for half_window in range(2, max_half_window, increment):
-        new_opening = grey_opening(y, [half_window * 2 + 1])
-        if relative_difference(opening, new_opening) < window_tol:
-            if hits == 0:
-                # keep just the first window that fits tolerance
-                best_half_window = half_window - increment
-            hits += 1
-            if hits >= max_hits:
-                half_window = best_half_window
-                break
-        elif hits:
-            hits = 0
-        opening = new_opening
-
-    return half_window
-
-
-def _setup_morphology(data, half_window=None, **window_kwargs):
-    """
-    Sets the starting parameters for morphology-based methods.
-
-    Parameters
-    ----------
-    data : array-like, shape (N,)
-        The y-values of the measured data, with N data points.
-    half_window : int, optional
-        The half-window used for the morphology functions. If a value is input,
-        then that value will be used. Default is None, which will optimize the
-        half-window size using pybaselines.morphological.optimize_window if
-        **window_kwargs has values, otherwise will use the number of data points
-        divided by 5.
-    **window_kwargs
-        Keyword arguments to pass to pybaselines.morphological.optimize_window.
-        Possible items are:
-
-            * 'increment': int
-                The step size for iterating half windows. Default is 1.
-            * 'max_hits': int
-                The number of consecutive half windows that must produce the same
-                morphological opening before accepting the half window as the
-                optimum value. Default is 1.
-            * 'window_tol': float
-                The tolerance value for considering two morphological openings as
-                equivalent. Default is 1e-6.
-            * 'max_half_window': int
-                The maximum allowable window size. If None (default), will be set
-                to (len(data) - 1) / 2.
-
-    Returns
-    -------
-    y : numpy.ndarray, shape (N,)
-        The y-values of the measured data, converted to a numpy array.
-    window : int
-        The accepted half window size.
-
-    Notes
-    -----
-    Ensures that window size is odd since morphological operations operate in
-    the range [-window, ..., window].
-
-    Half windows are dealt with rather than full window sizes to clarify their
-    usage. SciPy morphology operations deal with absolute window sizes.
-
-    """
-    y = np.asarray(data)
-    if half_window is not None:
-        window = half_window
-    elif window_kwargs:
-        window = optimize_window(y, **window_kwargs)
-    else:
-        window = y.shape[0] // 5
-
-    if window % 2 == 0:
-        window += 1
-
-    return y, window
+    x = (np.arange(0, 2 * window_size + 1) - window_size) / window_size
+    kernel = np.zeros_like(x)
+    # x[1:-1] is same as x[abs(x) < 1]
+    kernel[1:-1] = np.exp(-1 / (1 - (x[1:-1])**2))
+    return kernel / kernel.sum()
 
 
 def _avg_opening(y, half_window, opening=None):
@@ -203,7 +118,8 @@ def _avg_opening(y, half_window, opening=None):
     return 0.5 * (grey_dilation(opening, [window_size]) + grey_erosion(opening, [window_size]))
 
 
-def mpls(data, lam=1e6, p=0.0, order=2, tol=1e-3, max_iter=50, weights=None, **window_kwargs):
+def mpls(data, half_window=None, lam=1e6, p=0.0, diff_order=2, tol=1e-3, max_iter=50,
+         weights=None, **window_kwargs):
     """
     The Morphological penalized least squares (MPLS) baseline algorithm.
 
@@ -211,6 +127,10 @@ def mpls(data, lam=1e6, p=0.0, order=2, tol=1e-3, max_iter=50, weights=None, **w
     ----------
     data : array-like, shape (N,)
         The y-values of the measured data, with N data points.
+    half_window : int, optional
+        The half-window used for the morphology functions. If a value is input,
+        then that value will be used. Default is None, which will optimize the
+        half-window size using :func:`.optimize_window` and `window_kwargs`.
     lam : float, optional
         The smoothing parameter. Larger values will create smoother baselines.
         Default is 1e6.
@@ -218,9 +138,9 @@ def mpls(data, lam=1e6, p=0.0, order=2, tol=1e-3, max_iter=50, weights=None, **w
         The penalizing weighting factor. Must be between 0 and 1. Residuals
         above the data will be given p weight, and residuals below the data
         will be given p-1 weight. Default is 0.0.
-    order : {2, 1, 3}, optional
-        The order of the differential matrix. Default is 2 (second order
-        differential matrix).
+    diff_order : int, optional
+        The order of the differential matrix. Must be greater than 0. Default is 2
+        (second order differential matrix). Typical values are 2 or 1.
     max_iter : int, optional
         The max number of fit iterations. Default is 50.
     tol : float, optional
@@ -232,13 +152,6 @@ def mpls(data, lam=1e6, p=0.0, order=2, tol=1e-3, max_iter=50, weights=None, **w
         Values for setting the half window used for the morphology operations.
         Items include:
 
-            * 'half_window': int
-                The half-window used for the morphology functions. If a value is input,
-                then that value will be used. Default is None, which will optimize the
-                half-window size using pybaselines.morphological.optimize_window if
-                `increment`, `max_hits`, `window_tol`, or `max_half_window` are
-                specified, otherwise will use the number of data points
-                divided by 5.
             * 'increment': int
                 The step size for iterating half windows. Default is 1.
             * 'max_hits': int
@@ -251,12 +164,14 @@ def mpls(data, lam=1e6, p=0.0, order=2, tol=1e-3, max_iter=50, weights=None, **w
             * 'max_half_window': int
                 The maximum allowable window size. If None (default), will be set
                 to (len(data) - 1) / 2.
+            * 'min_half_window': int
+                The minimum half-window size. If None (default), will be set to 1.
 
     Returns
     -------
-    numpy.ndarray, shape (N,)
+    baseline : numpy.ndarray, shape (N,)
         The calculated baseline.
-    dict
+    params : dict
         A dictionary with the following items:
 
         * 'weights': numpy.ndarray, shape (N,)
@@ -264,24 +179,18 @@ def mpls(data, lam=1e6, p=0.0, order=2, tol=1e-3, max_iter=50, weights=None, **w
         * 'half_window': int
             The half window used for the morphological calculations.
 
-    Notes
-    -----
-    Although the MPLS is technically a penalized least squares algorithm,
-    it is included in pybaselines.morphological rather than
-    pybaselines.penalized_least_squares.
-
     References
     ----------
     .. [1] Li, Zhong, et al. Morphological weighted penalized least squares for
            background correction. Analyst, 2013, 138, 4483-4492.
 
     """
-    y, half_window = _setup_morphology(data, **window_kwargs)
+    y, half_wind = _setup_morphology(data, half_window, **window_kwargs)
     if weights is not None:
         w = weights
     else:
-        bkg = grey_opening(y, [2 * half_window + 1])
-        diff = np.diff(bkg, prepend=bkg[0], append=bkg[-1])
+        rough_baseline = grey_opening(y, [2 * half_wind + 1])
+        diff = np.diff(np.hstack([rough_baseline[0], rough_baseline, rough_baseline[-1]]))
         # diff == 0 means the point is on a flat segment, and diff != 0 means the
         # adjacent point is not the same flat segment. The union of the two finds
         # the endpoints of each segment, and np.flatnonzero converts the mask to
@@ -295,13 +204,24 @@ def mpls(data, lam=1e6, p=0.0, order=2, tol=1e-3, max_iter=50, weights=None, **w
             index = np.argmin(y[previous_segment:next_segment + 1]) + previous_segment
             w[index] = 1 - p
 
-    return (
-        _smooth(y, use_whittaker=True, lam=lam, order=order, weights=w),
-        {'weights': w, 'half_window': half_window}
+    using_pentapy = _HAS_PENTAPY and diff_order == 2
+    _, diagonals, weight_array = _setup_whittaker(
+        y, lam, diff_order, w, False, not using_pentapy, using_pentapy
     )
+    main_diag_idx = diff_order if using_pentapy else -1
+    diagonals[main_diag_idx] = diagonals[main_diag_idx] + weight_array
+    if using_pentapy:
+        baseline = _pentapy_solve(diagonals, weight_array * y, True, True, utils.PENTAPY_SOLVER)
+    else:
+        baseline = solveh_banded(
+            diagonals, weight_array * y, overwrite_ab=True, overwrite_b=True, check_finite=False
+        )
+
+    params = {'weights': weight_array, 'half_window': half_wind}
+    return baseline, params
 
 
-def mor(data, smooth=False, smooth_kwargs=None, **window_kwargs):
+def mor(data, half_window=None, **window_kwargs):
     """
     A Morphological based (Mor) baseline algorithm.
 
@@ -309,22 +229,14 @@ def mor(data, smooth=False, smooth_kwargs=None, **window_kwargs):
     ----------
     data : array-like, shape (N,)
         The y-values of the measured data, with N data points.
-    smooth : bool, optional
-        If True, will smooth the obtained baseline with Whittaker smoothing.
-        Default is False.
-    smooth_kwargs : dict, optional
-        A dictionary of arguments for smoothing if `smooth` is True. Default is None.
+    half_window : int, optional
+        The half-window used for the morphology functions. If a value is input,
+        then that value will be used. Default is None, which will optimize the
+        half-window size using :func:`.optimize_window` and `window_kwargs`.
     **window_kwargs
         Values for setting the half window used for the morphology operations.
         Items include:
 
-            * 'half_window': int
-                The half-window used for the morphology functions. If a value is input,
-                then that value will be used. Default is None, which will optimize the
-                half-window size using pybaselines.morphological.optimize_window if
-                `increment`, `max_hits`, `window_tol`, or `max_half_window` are
-                specified, otherwise will use the number of data points
-                divided by 5.
             * 'increment': int
                 The step size for iterating half windows. Default is 1.
             * 'max_hits': int
@@ -337,10 +249,12 @@ def mor(data, smooth=False, smooth_kwargs=None, **window_kwargs):
             * 'max_half_window': int
                 The maximum allowable window size. If None (default), will be set
                 to (len(data) - 1) / 2.
+            * 'min_half_window': int
+                The minimum half-window size. If None (default), will be set to 1.
 
     Returns
     -------
-    z : numpy.ndarray, shape (N,)
+    baseline : numpy.ndarray, shape (N,)
         The calculated baseline.
     dict
         A dictionary with the following items:
@@ -354,17 +268,14 @@ def mor(data, smooth=False, smooth_kwargs=None, **window_kwargs):
     Raman Spectra of Artistic Pigments. Applied Spectroscopy, 2010, 64, 595-600.
 
     """
-    y, half_window = _setup_morphology(data, **window_kwargs)
-    opening = grey_opening(y, [2 * half_window + 1])
-    z = np.minimum(opening, _avg_opening(y, half_window, opening))
-    if smooth:
-        smooth_kws = smooth_kwargs if smooth_kwargs is not None else {}
-        z = _smooth(z, **smooth_kws)
+    y, half_wind = _setup_morphology(data, half_window, **window_kwargs)
+    opening = grey_opening(y, [2 * half_wind + 1])
+    baseline = np.minimum(opening, _avg_opening(y, half_wind, opening))
 
-    return z, {'half_window': half_window}
+    return baseline, {'half_window': half_wind}
 
 
-def imor(data, tol=1e-3, max_iter=200, smooth=False, **window_kwargs):
+def imor(data, half_window=None, tol=1e-3, max_iter=200, **window_kwargs):
     """
     An Improved Morphological based (IMor) baseline algorithm.
 
@@ -372,24 +283,18 @@ def imor(data, tol=1e-3, max_iter=200, smooth=False, **window_kwargs):
     ----------
     data : array-like, shape (N,)
         The y-values of the measured data, with N data points.
+    half_window : int, optional
+        The half-window used for the morphology functions. If a value is input,
+        then that value will be used. Default is None, which will optimize the
+        half-window size using :func:`.optimize_window` and `window_kwargs`.
     tol : float, optional
         The exit criteria. Default is 1e-3.
     max_iter : int, optional
         The maximum number of iterations. Default is 200.
-    smooth : bool, optional
-        If True, will smooth the obtained baseline with mollification.
-        Default is False.
     **window_kwargs
         Values for setting the half window used for the morphology operations.
         Items include:
 
-            * 'half_window': int
-                The half-window used for the morphology functions. If a value is input,
-                then that value will be used. Default is None, which will optimize the
-                half-window size using pybaselines.morphological.optimize_window if
-                `increment`, `max_hits`, `window_tol`, or `max_half_window` are
-                specified, otherwise will use the number of data points
-                divided by 5.
             * 'increment': int
                 The step size for iterating half windows. Default is 1.
             * 'max_hits': int
@@ -402,10 +307,12 @@ def imor(data, tol=1e-3, max_iter=200, smooth=False, **window_kwargs):
             * 'max_half_window': int
                 The maximum allowable window size. If None (default), will be set
                 to (len(data) - 1) / 2.
+            * 'min_half_window': int
+                The minimum half-window size. If None (default), will be set to 1.
 
     Returns
     -------
-    z : numpy.ndarray, shape (N,)
+    baseline : numpy.ndarray, shape (N,)
         The calculated baseline.
     dict
         A dictionary with the following items:
@@ -413,60 +320,46 @@ def imor(data, tol=1e-3, max_iter=200, smooth=False, **window_kwargs):
         * 'half_window': int
             The half window used for the morphological calculations.
 
-    Notes
-    -----
-    Can optionally smooth the obtained background with mollification, which
-    makes the result more comparable to the iaMor algorith (Chen, H., et al.).
-
     References
     ----------
-    Dai, L., et al.. An Automated Baseline Correction Method Based on Iterative
+    Dai, L., et al. An Automated Baseline Correction Method Based on Iterative
     Morphological Operations. Applied Spectroscopy, 2018, 72(5), 731-739.
 
     """
-    y, half_window = _setup_morphology(data, **window_kwargs)
-    window_size = 2 * half_window + 1
-    if smooth:
-        kernel = np.array([
-            np.exp(-1 / (1 - (i / window_size)**2)) for i in range(-half_window + 1, half_window)
-        ])
-        kernel = kernel / kernel.sum()
-    z = np.minimum(y, _avg_opening(y, window_size))
+    y, half_wind = _setup_morphology(data, half_window, **window_kwargs)
+    baseline = np.minimum(y, _avg_opening(y, half_wind))
     for _ in range(max_iter - 1):
-        z_new = np.minimum(y, _avg_opening(z, half_window))
-        if smooth:
-            z_new = mollify(z_new, kernel)
-
-        if relative_difference(z, z_new) < tol:
+        baseline_new = np.minimum(y, _avg_opening(baseline, half_wind))
+        if relative_difference(baseline, baseline_new) < tol:
             break
-        z = z_new
+        baseline = baseline_new
 
-    return z, {'half_window': half_window}
+    return baseline, {'half_window': half_wind}
 
 
-def iamor(data, tol=1e-3, max_iter=200, **window_kwargs):
+def amormol(data, half_window=None, tol=1e-3, max_iter=200, pad_kwargs=None, **window_kwargs):
     """
-    Iteratively averaging morphological (iaMor) baseline.
+    Iteratively averaging morphological and mollified (aMorMol) baseline.
 
     Parameters
     ----------
     data : array-like, shape (N,)
         The y-values of the measured data, with N data points.
+    half_window : int, optional
+        The half-window used for the morphology functions. If a value is input,
+        then that value will be used. Default is None, which will optimize the
+        half-window size using :func:`.optimize_window` and `window_kwargs`.
     tol : float, optional
         The exit criteria. Default is 1e-3.
     max_iter : int, optional
         The maximum number of iterations. Default is 200.
+    pad_kwargs : dict, optional
+        A dictionary of keyword arguments to pass to :func:`.pad_edges` for
+        padding the edges of the data to prevent edge effects from convolution.
     **window_kwargs
         Values for setting the half window used for the morphology operations.
         Items include:
 
-            * 'half_window': int
-                The half-window used for the morphology functions. If a value is input,
-                then that value will be used. Default is None, which will optimize the
-                half-window size using pybaselines.morphological.optimize_window if
-                `increment`, `max_hits`, `window_tol`, or `max_half_window` are
-                specified, otherwise will use the number of data points
-                divided by 5.
             * 'increment': int
                 The step size for iterating half windows. Default is 1.
             * 'max_hits': int
@@ -479,10 +372,12 @@ def iamor(data, tol=1e-3, max_iter=200, **window_kwargs):
             * 'max_half_window': int
                 The maximum allowable window size. If None (default), will be set
                 to (len(data) - 1) / 2.
+            * 'min_half_window': int
+                The minimum half-window size. If None (default), will be set to 1.
 
     Returns
     -------
-    z : numpy.ndarray, shape (N,)
+    baseline : numpy.ndarray, shape (N,)
         The calculated baseline.
     dict
         A dictionary with the following items:
@@ -490,10 +385,6 @@ def iamor(data, tol=1e-3, max_iter=200, **window_kwargs):
         * 'half_window': int
             The half window used for the morphological calculations.
 
-    Notes
-    -----
-    There is no established name for this method, so just used iaMor.
-
     References
     ----------
     Chen, H., et al. An Adaptive and Fully Automated Baseline Correction
@@ -501,27 +392,283 @@ def iamor(data, tol=1e-3, max_iter=200, **window_kwargs):
     Mollifications. Applied Spectroscopy, 2019, 73(3), 284-293.
 
     """
-    y, half_window = _setup_morphology(data, **window_kwargs)
-    window_size = 2 * half_window + 1
-    kernel = np.array([
-        np.exp(-1 / (1 - (i / window_size)**2)) for i in range(-half_window + 1, half_window)
-    ])
-    # normalize the kernel before convolution so that can just do
-    # np.convolve(y, normalized_kernel) rather than
-    # np.convolve(y, kernel) / np.convolve(np.ones(y.shape[0]), kernel)
-    kernel = kernel / kernel.sum()
+    y, half_wind = _setup_morphology(data, half_window, **window_kwargs)
+    window_size = 2 * half_wind + 1
+    kernel = _mollifier_kernel(window_size)
+    data_bounds = slice(window_size, -window_size)
 
-    z = y
+    pad_kws = pad_kwargs if pad_kwargs is not None else {}
+    y = pad_edges(y, window_size, **pad_kws)
+    baseline = y
     for _ in range(max_iter):
-        z_new = mollify(
+        baseline_new = padded_convolve(
             np.minimum(
                 y,
-                0.5 * (grey_closing(z, [window_size]) + grey_opening(z, [window_size]))
+                (grey_closing(baseline, [window_size]) + grey_opening(baseline, [window_size])) / 2
             ),
             kernel
         )
-        if relative_difference(z, z_new) < tol:
+        if relative_difference(baseline[data_bounds], baseline_new[data_bounds]) < tol:
             break
-        z = z_new
+        baseline = baseline_new
 
-    return z, {'half_window': half_window}
+    return baseline[data_bounds], {'half_window': half_wind}
+
+
+def mormol(data, half_window=None, tol=1e-3, max_iter=250, smooth_half_window=None,
+           pad_kwargs=None, **window_kwargs):
+    """
+    Iterative morphological and mollified (MorMol) baseline.
+
+    Parameters
+    ----------
+    data : array-like, shape (N,)
+        The y-values of the measured data, with N data points.
+    half_window : int, optional
+        The half-window used for the morphology functions. If a value is input,
+        then that value will be used. Default is None, which will optimize the
+        half-window size using :func:`.optimize_window` and `window_kwargs`.
+    tol : float, optional
+        The exit criteria. Default is 1e-3.
+    max_iter : int, optional
+        The maximum number of iterations. Default is 200.
+    smooth_half_window : int, optional
+        The half-window to use for smoothing the data before performing the
+        morphological operation. Default is None, which will use a value of 1,
+        which gives no smoothing.
+    pad_kwargs : dict, optional
+        A dictionary of keyword arguments to pass to :func:`.pad_edges` for
+        padding the edges of the data to prevent edge effects from convolution.
+    **window_kwargs
+        Values for setting the half window used for the morphology operations.
+        Items include:
+
+            * 'increment': int
+                The step size for iterating half windows. Default is 1.
+            * 'max_hits': int
+                The number of consecutive half windows that must produce the same
+                morphological opening before accepting the half window as the
+                optimum value. Default is 1.
+            * 'window_tol': float
+                The tolerance value for considering two morphological openings as
+                equivalent. Default is 1e-6.
+            * 'max_half_window': int
+                The maximum allowable window size. If None (default), will be set
+                to (len(data) - 1) / 2.
+            * 'min_half_window': int
+                The minimum half-window size. If None (default), will be set to 1.
+
+    Returns
+    -------
+    baseline : numpy.ndarray, shape (N,)
+        The calculated baseline.
+    dict
+        A dictionary with the following items:
+
+        * 'half_window': int
+            The half window used for the morphological calculations.
+
+    References
+    ----------
+    Koch, M., et al. Iterative morphological and mollifier-based baseline
+    correction for Raman spectra. J Raman Spectroscopy, 2017, 48(2), 336-342.
+
+    """
+    y, half_wind = _setup_morphology(data, half_window, **window_kwargs)
+    window_size = 2 * half_wind + 1
+    kernel = _mollifier_kernel(window_size)
+    if smooth_half_window is None:
+        smooth_half_window = 1
+    smooth_kernel = _mollifier_kernel(smooth_half_window)
+    data_bounds = slice(window_size, -window_size)
+
+    pad_kws = pad_kwargs if pad_kwargs is not None else {}
+    y = pad_edges(y, window_size, **pad_kws)
+    baseline = np.zeros(y.shape[0])
+    for _ in range(max_iter):
+        y_smooth = padded_convolve(y - baseline, smooth_kernel)
+        baseline_new = baseline + padded_convolve(grey_erosion(y_smooth, window_size), kernel)
+        if relative_difference(baseline[data_bounds], baseline_new[data_bounds]) < tol:
+            break
+        baseline = baseline_new
+
+    return baseline[data_bounds], {'half_window': half_wind}
+
+
+def _changing_rolling_ball(y, half_window):
+    """
+    Calculates the rolling ball algorithm with a changing window size.
+
+    Parameters
+    ----------
+    y : array-like, shape (N,)
+        The y-values of the measured data, with N data points.
+    half_window : array-like, shape (N,)
+        The array of half-windows to use for the morphology functions.
+
+    Returns
+    -------
+    rough_baseline : numpy.ndarray, shape (N,)
+        The array of the baseline after the morphology functions.
+
+    Raises
+    ------
+    ValueError
+        Raised if y and half_window do not have the same length.
+
+    References
+    ----------
+    Kneen, M.A., et al. Algorithm for fitting XRF, SEM and PIXE X-ray spectra
+    backgrounds. Nuclear Instruments and Methods in Physics Research B, 1996,
+    109, 209-213.
+
+    """
+    num_y = len(y)
+    if len(half_window) != num_y:
+        raise ValueError('half_window array must be the same size as the data array')
+
+    minimum_y = np.empty(num_y)
+    for i, half_win in enumerate(half_window):
+        minimum_y[i] = min(y[max(0, i - half_win):min(i + half_win + 1, num_y)])
+
+    rough_baseline = np.empty(num_y)
+    for i, half_win in enumerate(half_window):
+        rough_baseline[i] = max(minimum_y[max(0, i - half_win):min(i + half_win + 1, num_y)])
+
+    return rough_baseline
+
+
+def _changing_smooth_window(rough_baseline, smooth_half_window):
+    """
+    Smooths an array with the rolling average and a changing window size.
+
+    Parameters
+    ----------
+    rough_baseline : numpy.ndarray, shape (N,)
+        The array of the baseline before smoothing, with N data points.
+    smooth_half_window : array-like, shape (N,)
+        The array of half-windows to use for smoothing.
+
+    Returns
+    -------
+    baseline : numpy.ndarray, shape (N,)
+        The smoothed baseline array.
+
+    Raises
+    ------
+    ValueError
+        Raised if rough_baseline and smooth_half_window do not have the same length.
+
+    References
+    ----------
+    Kneen, M.A., et al. Algorithm for fitting XRF, SEM and PIXE X-ray spectra
+    backgrounds. Nuclear Instruments and Methods in Physics Research B, 1996,
+    109, 209-213.
+
+    """
+    num_y = rough_baseline.shape[0]
+    if len(smooth_half_window) != num_y:
+        raise ValueError('smooth_half_window array must be the same size as the data array')
+
+    baseline = np.empty(num_y)
+    for i, half_win in enumerate(smooth_half_window):
+        baseline_slice = rough_baseline[max(0, i - half_win):min(i + half_win + 1, num_y)]
+        baseline[i] = baseline_slice.sum() / baseline_slice.size
+
+    return baseline
+
+
+def rolling_ball(data, half_window=None, smooth_half_window=None, pad_kwargs=None, **window_kwargs):
+    """
+    The rolling ball baseline algorithm.
+
+    Applies a minimum and then maximum moving window, and subsequently smooths the
+    result, giving a baseline that resembles rolling a ball across the data.
+
+    Parameters
+    ----------
+    data : array-like, shape (N,)
+        The y-values of the measured data, with N data points.
+    half_window : int or array-like(int), optional
+        The half-window used for the morphology functions. If a value is input,
+        then that value will be used. Default is None, which will optimize the
+        half-window size using :func:`.optimize_window` and `window_kwargs`. If
+        an array of integers is input, its length must be equal to N.
+    smooth_half_window : int or array-like(int), optional
+        The half-window to use for smoothing the data after performing the
+        morphological operation. Default is None, which will use the same
+        value as used for the morphological operation. If an array of integers
+        is input, its length must be equal to N.
+    pad_kwargs : dict, optional
+        A dictionary of keyword arguments to pass to :func:`.pad_edges` for
+        padding the edges of the data to prevent edge effects from the moving average.
+    **window_kwargs
+        Values for setting the half window used for the morphology operations.
+        Items include:
+
+            * 'increment': int
+                The step size for iterating half windows. Default is 1.
+            * 'max_hits': int
+                The number of consecutive half windows that must produce the same
+                morphological opening before accepting the half window as the
+                optimum value. Default is 1.
+            * 'window_tol': float
+                The tolerance value for considering two morphological openings as
+                equivalent. Default is 1e-6.
+            * 'max_half_window': int
+                The maximum allowable window size. If None (default), will be set
+                to (len(data) - 1) / 2.
+            * 'min_half_window': int
+                The minimum half-window size. If None (default), will be set to 1.
+
+    Returns
+    -------
+    baseline : numpy.ndarray, shape (N,)
+        The calculated baseline.
+    dict
+        A dictionary with the following items:
+
+        * 'half_window': int or numpy.ndarray(int)
+            The half window or array of half windows used for the
+            morphological calculations.
+
+    Notes
+    -----
+    To use a changing window size for either the morphological or smoothing
+    operations, the half windows must be arrays. Otherwise, the size of the
+    rolling ball is assumed to be constant.
+
+    References
+    ----------
+    Kneen, M.A., et al. Algorithm for fitting XRF, SEM and PIXE X-ray spectra
+    backgrounds. Nuclear Instruments and Methods in Physics Research B, 1996,
+    109, 209-213.
+
+    Liland, K., et al. Optimal Choice of Baseline Correction for Multivariate
+    Calibration of Spectra. Applied Spectroscopy, 2010, 64(9), 1007-1016.
+
+    """
+    if half_window is None or isinstance(half_window, int):
+        y, half_wind = _setup_morphology(data, half_window, **window_kwargs)
+        rough_baseline = grey_opening(y, 2 * half_wind + 1)
+    else:
+        # do not need to convert data to numpy array
+        half_wind = half_window
+        rough_baseline = _changing_rolling_ball(data, half_wind)
+
+    if smooth_half_window is None:
+        smooth_half_window = half_wind
+
+    if isinstance(smooth_half_window, int):
+        if smooth_half_window == 0:
+            baseline = rough_baseline
+        else:
+            pad_kws = pad_kwargs if pad_kwargs is not None else {}
+            baseline = uniform_filter1d(
+                pad_edges(rough_baseline, smooth_half_window, **pad_kws),
+                2 * smooth_half_window + 1
+            )[smooth_half_window:-smooth_half_window]
+    else:
+        baseline = _changing_smooth_window(rough_baseline, smooth_half_window)
+
+    return baseline, {'half_window': half_wind}
